// Copyright (c) 2014, The Shelf REST project authors.
// Please see the AUTHORS file for details.
// All rights reserved. Use of this source code is governed by
// a BSD 2-Clause License that can be found in the LICENSE file.

library shelf_rest.hateoas;

import 'src/preconditions.dart';
import 'package:option/option.dart';
import 'package:uri/uri.dart';
import 'package:path/path.dart' as path;

final _p = path.url;

class Link {
  final String rel;
  final String href;
  Uri get hrefAsUri => Uri.parse(href);
  UriTemplate get hrefAsUriTemplate => new UriTemplate(href);

  Link(String rel, String href)
      : this.rel = checkNotNull(rel),
        this.href = checkNotNull(href);

  Link.uri(String rel, Uri href)
      : this(rel, href.toString());

  Link.uriTemplate(String rel, UriTemplate href)
      : this(rel, href.template);

  Link.self(Uri href) : this.uri('self', href);

  static Option<Link> optUri(String rel, Option<Uri> hrefOpt) =>
      hrefOpt.map((href) => new Link.uri(rel, href));

  static Option<Link> optUriTemplate(String rel, Option<UriTemplate> hrefOpt) =>
      hrefOpt.map((href) => new Link.uriTemplate(rel, href));

  Link.fromJson(Map<String, Object> json)
      : this(json['rel'], json['href']);

  Map<String, Object> toJson() => {
    'rel': rel,
    'href': href
  };

  @override
  int get hashCode => href.hashCode;

  @override
  bool operator==(other) =>
      other is Link && other.rel == rel && other.href == href;

  @override
  String toString() => 'Link($rel : $href)';
}

abstract class Links {
  List<Link> get links;

  Map toJson() => {
      'links': links.map((l) => l.toJson()).toList(growable: false)
  };

  void addJson(Map parentJson) => parentJson.addAll(toJson());
}

class DefaultLinks extends Links {
  final List<Link> links;

  DefaultLinks(this.links);
}

abstract class ResourceModel<T> {
  final ResourceLinks links;

  ResourceModel(this.links);

  Map toJson() => links.toJson();
}

abstract class Identified {
  String get id;
}

class SearchResult<R> {
  final int totalCount;
  final int page;
  final List<R> results;
  final int numPages;
  int get pageSize => numPages > 0 ? (totalCount / numPages).ceil() : 1;

  SearchResult(int totalCount, int numPages, int page,
      this.results)
      : this.totalCount = checkNotNull(totalCount),
        this.numPages = checkNotNull(numPages),
        this.page = checkNotNull(page) {
    ensure(totalCount, greaterThanOrEqualTo(0));
    ensure(numPages, greaterThanOrEqualTo(0));
    ensure(page, greaterThanOrEqualTo(0));
  }

  SearchResult.fromJson(Map json, R resourceFromJson(Map resourceJson))
      : this(
          json['totalCount'],
          json['numPages'],
          json['page'],
          (json['results'] as Iterable).map(resourceFromJson)
            .toList(growable: false)
          );


  Map toJson() => {
    'totalCount': totalCount,
    'numPages': numPages,
    'page': page,
    'results': results.map((r) => r.toJson()).toList(growable: false)
  };
}

class SearchResourceModel<R> extends SearchResult<R> {
  final List<Link> links; // TODO: change to Links??
  final Uri self;
  final Option<Uri> previousPage;
  final Option<Uri> nextPage;

  SearchResourceModel(int totalCount, int numPages, int page, List<R> results,
      Uri self, { Uri previousPage, Uri nextPage })
      : this.self = checkNotNull(self),
        this.previousPage = new Option(previousPage),
        this.nextPage = new Option(nextPage),
        links = (new LinkHelper(self)
          ..addLink('previousPage', previousPage)
          ..addLink('nextPage', nextPage))
          .links,
        super(totalCount, numPages, page, results);

  SearchResourceModel.fromResult(SearchResult<R> result,
      Uri self, { Uri previousPage, Uri nextPage })
      : this(result.totalCount, result.numPages, result.page, result.results,
          self, previousPage: previousPage, nextPage: nextPage);

  factory SearchResourceModel.fromJson(Map json,
      R resourceFromJson(Map resourceJson)) {

    final helper = new LinkJsonHelper(json);

    return new SearchResourceModel.fromResult(
        new SearchResult.fromJson(json, resourceFromJson),
            helper.uri('self'),
            previousPage: helper.uri('previousPage'),
            nextPage: helper.uri('nextPage'));
  }

  Map toJson() =>
    super.toJson()..addAll({
      'links': links.map((l) => l.toJson()).toList(growable: false)
  });
}



class LinkJsonHelper {
  List<Link> links;

  LinkJsonHelper(Map json) {
    final linkJson = (json['links'] as List);

    links = linkJson.map((link) =>
      new Link.fromJson(link)).toList();
  }

  _link(String rel, convert(Link link)) {
    final result = links.where((l) => l.rel == rel);
    return result.isEmpty ? null : convert(result.first);
  }

  Uri uri(String rel) => _link(rel, (link) => link.hrefAsUri);

  UriTemplate uriTemplate(String rel) =>
      _link(rel, (link) => link.hrefAsUriTemplate);

}


class LinkHelper {
  final List<Link> links;

  LinkHelper([self]) :
    links = self != null ? [new Link.self(self)] : [];

  void addLink(String rel, href) {
    if (href != null) {
      links.add(new Link(rel, href.toString()));
    }
  }

}

//class SearchLinks extends Links {
//  Uri self; // ?foo=bar
//  List<InstanceLinks>
//  // search different as contains links for each result
//}

typedef ResourceLinks ResourceLinksFactory(resourceId);

class ResourceLinks extends Links {
  final Uri self;
  final Option<Uri> delete;
  final Option<Uri> update;
  final Map<dynamic, ChildResourceLinks> childResourceLinks;

  List<Link> get links {
    final l =
      [new Some(new Link.self(self)),
       Link.optUri('delete', delete),
       Link.optUri('update', update)]
    ..addAll(_childLinks);

    return l.where((o) => o is Some).map((o) => o.get()).toList(growable: false);
  }

  Iterable<Option<Link>> get _childLinks =>
      childResourceLinks.keys.expand((k) =>
          [ childSearchLink(k), childFindLink(k), childCreateLink(k),
            childUpdateLink(k), childDeleteLink(k)]);

  ResourceLinks(this.self, this.delete, this.update, this.childResourceLinks);

  ResourceLinks.nullable(Uri self, Uri delete, Uri update,  Map<dynamic,
      ChildResourceLinks> childResourceLinks)
      : this(self, new Option(delete), new Option(update), childResourceLinks);

  Option<UriTemplate> childSearch(dynamic path) =>
      new Option(childResourceLinks[path]).flatMap((crl) => crl.search);

  Option<UriTemplate> childFind(dynamic path) =>
      new Option(childResourceLinks[path]).flatMap((crl) => crl.find);

  Option<Uri> childCreate(dynamic path) =>
      new Option(childResourceLinks[path]).flatMap((crl) => crl.create);

  Option<UriTemplate> childUpdate(dynamic path) =>
      new Option(childResourceLinks[path]).flatMap((crl) => crl.update);

  Option<UriTemplate> childDelete(dynamic path) =>
      new Option(childResourceLinks[path]).flatMap((crl) => crl.delete);

  Option<Link> childSearchLink(path) =>
      Link.optUriTemplate(_pathToName(path, 'search'), childSearch(path));

  Option<Link> childFindLink(path) =>
      Link.optUriTemplate(_pathToName(path, 'find'), childFind(path));

  Option<Link> childCreateLink(path) =>
      Link.optUri(_pathToName(path, 'create'), childCreate(path));

  Option<Link> childUpdateLink(path) =>
      Link.optUriTemplate(_pathToName(path, 'update'), childUpdate(path));

  Option<Link> childDeleteLink(path) =>
      Link.optUriTemplate(_pathToName(path, 'delete'), childDelete(path));
}

class ChildResourceLinks  {
  final Option<UriTemplate> search; // template: <parentId>/childPath/<search params>
  final Option<UriTemplate> find; // template: <parentId>/childPath/{child path param}
  final Option<Uri> create; // template: <parentId>/childPath/{child path param}
                              // or <parentId>/childPath for POST
  final Option<UriTemplate> update; // template: <parentId>/childPath/{child path param}
  final Option<UriTemplate> delete; // template: <parentId>/childPath/{child path param}

  // TODO: adhoc needs to fit in some way unless we leave it to user to create
//  List<UriTemplate> adhocTemplates;
//  List<Uri> adhocUris;

//  List<Link> get links => [];

  ChildResourceLinks(this.search, this.find, this.create, this.update,
      this.delete);

  ChildResourceLinks.nullable(UriTemplate search, UriTemplate find,
      Uri create, UriTemplate update, UriTemplate delete)
      : this(new Option(search), new Option(find), new Option(create),
          new Option(update), new Option(delete));

}


String _pathToName(String path, String qualifier) {
  final segs = _p.split(path);
  final capSegs = segs.where((s) => s != _p.separator).map(_capitalise);
  return ([qualifier]..addAll(capSegs)).join('');
}


String _capitalise(String s) =>
    s.substring(0, 1).toUpperCase() + s.substring(1);

